/* Laser scanner controller

   Copyright (C) 2013 Frederic Rible (frible@teaser.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 3 of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the program; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.
 */

#include <EEPROM.h>
#include <AFMotor.h>

AF_Stepper motor(200, 1);

int MaxPos = 5000;
int Homed = 0;

int CurrentPos = 0;
int PredefinedPos = 0;
int LaserPin = 10;
int LightPin = 9;
int SwitchPin = 13;
unsigned long MotorReleaseTime = 0;
#define MOTOR_RELEASE_DELAY 1000

#define MODE_CAL 1
#define MODE_SCAN 2
#define MODE_TEXTURE 3
int CurrentDavidMode = MODE_CAL;
boolean AutoStop = false;
int EndOfScanPos = 0;

int EEPROM_read_int(int addr)
{
  return EEPROM.read(addr) + (EEPROM.read(addr+1) << 8);
}

void EEPROM_write_int(int addr, int value)
{
  EEPROM.write(addr, value & 255);
  EEPROM.write(addr+1, value >> 8);
}  

void setup() {
  pinMode(LaserPin, OUTPUT);
  pinMode(LightPin, OUTPUT);
  pinMode(SwitchPin, INPUT_PULLUP);
  Serial.begin(115200);
  motor.setSpeed(100);
  PredefinedPos = EEPROM_read_int(0);
  EndOfScanPos = EEPROM_read_int(2);
}

void RelativeMove(int v)
{
  if (Homed) {
    if (CurrentPos + v < 0) v = -CurrentPos;
    if (CurrentPos + v > MaxPos) v = MaxPos - CurrentPos;
  }

  if (v >= 0) motor.step(v, BACKWARD, SINGLE);
  else motor.step(-v, FORWARD, SINGLE);
  CurrentPos += v;
  MotorReleaseTime = millis() + MOTOR_RELEASE_DELAY;
  //Serial.println(CurrentPos);
}

void AbsoluteMove(int v)
{
  RelativeMove(v - CurrentPos);
}

void Home()
{
  int count;
  Homed = 0;

  // Move out homing switch
  for (count=250; count>0; count--) {
    if (!digitalRead(SwitchPin)) break;
    RelativeMove(+1);
  }

  if (count == 0) {
    //Serial.println("HOME ERROR: swith always active");
    return;
  }

  //  Move toward homing switch
  for (count=MaxPos+100; count>0; count--) {
    if (digitalRead(SwitchPin)) break;
    RelativeMove(-1);
  }

  if (count == 0) {
    //Serial.println("HOME ERROR: cannot hit switch");
    return;
  }

  CurrentPos = 0;
  Homed = 1;
}

void SwitchLight(int v)
{
  digitalWrite(LightPin, v);
}

void SwitchLaser(int v)
{
  digitalWrite(LaserPin, v);
}

void ProcessCommand()
{
  char buf[32];
  int v;
  boolean minus = false;
  int count;
  int i;

  int c = Serial.read();
  if (!Homed || c == 'O') {
    SwitchLight(0);
    SwitchLaser(0);
    Home();
  }

  if (!Homed) {
    Serial.print('T');  // Stop scan
    return;
  }

  switch (c) {
  case '+':
    RelativeMove(-100);
    break;
  case '-':
    RelativeMove(100); 
    break;
  case 'm':
    count  = Serial.readBytesUntil(';', buf, sizeof(buf));
    if (count <= 0) break;
    v = 0;
    switch (buf[0]) {
    case '+':
      i = 1;
      break;
    case '-':
      i = 1;
      minus = true;
      break;
    default:
      i = 0;
      break;
    }
    while (i < count) {
      v = 10*v + (buf[i] - '0');
      i++;
    }  
    if (minus) v = -v;
    RelativeMove(-v);
    if (AutoStop && (CurrentPos == EndOfScanPos)) {  // Auto stop ?
      Serial.println('T');
    }
    break;
  case 'P':  // Go to predefined position
    AbsoluteMove(PredefinedPos);
    break;    
  case '0':  // Set initial position
    PredefinedPos = CurrentPos;
    EEPROM_write_int(0, PredefinedPos);
    break;
  case '1':  // Calibration mode
    CurrentDavidMode = MODE_CAL;
    SwitchLight(1);
    break; 
  case '2':  // Scan mode
    SwitchLight(0);
    if (CurrentDavidMode != MODE_SCAN)  AbsoluteMove(PredefinedPos);
    CurrentDavidMode = MODE_SCAN;
    break;
  case '3':  // Texture mode
    CurrentDavidMode = MODE_TEXTURE;
    SwitchLight(1);
    break;
  case 'l':  // Laser off
    SwitchLaser(0);
    break;
  case 'L':  // Laser On
    SwitchLaser(1);
    break;
  case 'R':  // Start scanning repetion
    SwitchLight(0);
    SwitchLaser(1);
    AutoStop = true;
    break;
  case 'S':  // Start scanning record
    SwitchLight(0);
    SwitchLaser(1);
    AutoStop = false;
    break;
  case 'T': // Stop scanning
    SwitchLaser(0);
    if (!AutoStop) {  // Record the end of scan position for repetition autostop
      EndOfScanPos = CurrentPos;
      EEPROM_write_int(2, EndOfScanPos);
    }
    AutoStop = false;
    if (CurrentDavidMode == MODE_SCAN) AbsoluteMove(PredefinedPos);
    break;
  case 'O':  // Open => Homing
    break;
  case 'C':  // Close
    SwitchLight(0);
    SwitchLaser(0);
    break;    
  }
}

void loop() {
  if (Serial.available() > 0) {
    ProcessCommand();
  }
  if (MotorReleaseTime && (millis() > MotorReleaseTime)) {
    motor.release();
    MotorReleaseTime = 0;
  }
}

